Unit Atol_Driver;

{$mode objfpc}{$H+} 

Interface

Uses
  Classes, SysUtils, synaser, LConvEncoding;

Const
  sENQ = #$05; //Запрос
  sACK = #$06; //Подтверждение
  sSTX = #$02; //Начало текста
  sETX = #$03; //Конец текста
  sEOT = #$04; //Конец передачи
  sNAK = #$15; //Отрицание
  sDLE = #$10; //Экранирование управляющих символов

  timeout_T1 = 1000;
  timeout_T2 = 2000;
  timeout_T3 = 1000;
  timeout_T4 = 1000;
  timeout_T5 = 10000;
  timeout_T6 = 1000;
  timeout_T7 = 1000;

  BMax = 1024;

  AtolModeCount = 6;
  AtolModeNames: Array [0..AtolModeCount] Of String =
    ('Выбор',
    'Режим регистрации',
    'Режим отчетов без гашения',
    'Режим отчетов с гашением',
    'Режим программирования',
    'Режим доступа к ФП',
    'Режим доступа к ЭКЛЗ');

  KKM_CODEPAGE: Array[0..254] Of String =
    ('А', 'Б', 'В', 'Г', 'Д', 'Е', 'Ж', 'З', 'И', 'Й', 'К', 'Л', 'М', 'Н', 'О', 'П', 'Р', 'С', 'Т', 'У',
    'Ф', 'Х', 'Ц', 'Ч', 'Ш', 'Щ', 'Ъ', 'Ы', 'Ь', 'Э', 'Ю', 'Я', ' ', '!', '"', '#', '№', '%', '&', '’',
    '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';',
    '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\', ']', '^', '_', '`', 'a', 'b', 'c',
    'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
    'x', 'y', 'z', '{', '|', '}', '~', ' ', 'а', 'б', 'в', 'г', 'д', 'е', 'ж', 'з', 'и', 'й', 'к', 'л',
    'м', 'н', 'о', 'п', 'р', 'с', 'т', 'у', 'ф', 'х', 'ц', 'ч', 'ш', 'щ', 'ъ', 'ы', 'ь', 'э', 'ю', 'я',
    '$', '€', '—', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
    '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
    '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
    '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
    '', '', '', '', '', '', '', '', '', '', '', '', '', '', '»');

Type
  TDrvDebugFunc = Procedure(msg: String) Of Object;

  TFRAtolDriver = Class(TObject)
  Private
    FAccessPassword:           Integer;
    FDataBits:                 Integer;
    FFieldNumber:              Integer;
    FFieldType:                Boolean;
    FParitySim:                Char;
    FRowNumber:                Integer;
    FTableNumber:              Integer;
    FXonXOff:                  Boolean;
    FRTSCTS:                   Boolean;
    FStopBits:                 Integer;
    FPortName:                 String;
    FPortSpeed:                Integer;
    FSerialPort:               TBlockSerial;
    FTestMode:                 Boolean;
    FTypeClose:                Byte;
    FCheckType:                Byte;
    FMaxLineWidth:             Byte;
    FDeviceNumber:             Integer;
    FBatteryLow:               Boolean;
    FBufferPDEmpty:            Boolean;
    FBufferPDFull:             Boolean;
    FCaption:                  String;
    FCheckNumber:              Integer;
    FCheckState:               Byte;
    FCount:                    Integer;
    FCoverOpened:              Boolean;
    FPrinterCutMechanismError: Boolean;
    FDepartment:               Integer;
    FPrinterOverheatError:     Boolean;
    FECRDateTime:              TDateTime;
    FPrintConected:            Boolean;
    FKKMConnected:             Boolean;
    FOutOfPaper:               Boolean;
    FPaperError:               Boolean;
    FPrice:                    Currency;
    FQuantity:                 Double;
    FSession:                  Integer;
    FErrorCode:                Integer;
    FErrorCodeStr:             String;
    FMode:                     Byte;
    FStatus:                   Byte;
    FPassword:                 String;
    FSessionOpened:            Boolean;
    FSubMode:                  Byte;
    FSumm:                     Currency;
    FTextWrap:                 Integer;
    FUmodel:                   Byte;
    FReportType:               Byte;
    FDebugFunc:                TDrvDebugFunc;
    Procedure SetErrorCode(AErrorCode: Integer);
    Procedure SetDepartment(AValue: Integer);
    Procedure SetCheckType(AValue: Byte);
    Procedure SetCount(AValue: Integer);
    Procedure SetAccessPassword(AValue: Integer);
    Procedure SetCaption(AValue: String);
    Function GetCaption: String;
    Procedure SetPassword(AValue: String);
    Procedure SetQuantity(AValue: Double);
    Procedure SetTestMode(AValue: Boolean);
    Procedure SetTypeClose(AValue: Byte);
    Procedure SetKKMConected(AValue: Boolean);
    Procedure SetParity(AValue: Char);
    Procedure SetPortName(AValue: String);
    Procedure SetPortSpeed(AValue: Integer);
    Procedure SetPrice(AValue: Currency);
    Procedure Set_FieldType(AValue: Boolean);
  Protected
    Procedure DebugMsg(msg: String);
    Procedure InternalGetStatus;
    Procedure InternalSetMode(AMode: Byte);
    Function MakeCommand(AData: Ansistring): Ansistring;
    Procedure SendCommand(cmd: Ansistring; TimeOutAnsverCmd: Integer = timeout_T5);
    Function ReadAnsver: Ansistring;
    Procedure ReportX;             // Вывод X-отчета
    Procedure ReportZ;             // Начало вывода Z-отчета
  Public
    Constructor Create;
    Destructor Destroy; Override;

    Procedure Beep;                              // Гудок
    Procedure CancelCheck;                       // Отмена текущего чека
    Procedure CashIncome;                        // Внесение наличных
    Procedure CashOutcome;                       // Изъятие наличных
    Procedure CloseCheck;                        // Закрыть чек
    Procedure FullCut;                           // Полная отрезка ленты
    Procedure GetDeviceMetrics;                  // Получить параметры ККМ
    Procedure GetStatus;                         // Получить состояние ККМ
    Procedure NewDocument;                       // Создание нового документа
    Procedure OpenCheck;                         // Открыть чек
    Procedure OpenDrawer;                        // Открыть денежный ящик
    Procedure OpenSession;                       // Открыть смену
    Procedure PartialCut;                        // Частичная отрезка ленты
    Procedure Payment;                           // Продажа
    Procedure PrintHeader;                       // Печать клише
    Procedure PrintString;                       // Печать строки
    Procedure Registration;                      // Регистрация
    Procedure ResetMode;                         // Выйти из режима
    Procedure Return;                            // Возврат
    Function SetMode(AValue: Byte): Integer;     // Установить режим
    Procedure Report;                            // Снятие отчетов
    Procedure ReadTable;                         // Чтение таблицы
    Procedure WriteTable;                        // Запись таблицы

    Property DeviceEnabled: Boolean Read FKKMConnected Write SetKKMConected;
    Property DataBits: Integer Read FDataBits Write FDataBits;
    Property Parity: Char Read FParitySim Write SetParity;
    Property XOnXOff: Boolean Read FXOnXOff Write FXOnXOff;
    Property RTSCTS: Boolean Read FRTSCTS Write FRTSCTS;
    Property StopBits: Integer Read FStopBits Write FStopBits;
    Property ResultCode: Integer Read FErrorCode;
    Property ResultDescription: String Read FErrorCodeStr;
    Property Department: Integer Read FDepartment Write SetDepartment;
    Property CheckType: Byte Read FCheckType Write SetCheckType;
    Property Count: Integer Read FCount Write SetCount;
    Property AccessPassword: Integer Read FAccessPassword Write SetAccessPassword;
    Property Caption: String Read GetCaption Write SetCaption;
    Property Name: String Read FCaption Write SetCaption;
    Property Password: String Read FPassword Write SetPassword;
    Property PortName: String Read FPortName Write SetPortName;
    Property PortSpeed: Integer Read FPortSpeed Write SetPortSpeed;
    Property Price: Currency Read FPrice Write SetPrice;
    Property Quantity: Double Read FQuantity Write SetQuantity;
    Property Summ: Currency Read FSumm Write FSumm;
    Property TestMode: Boolean Read FTestMode Write SetTestMode;
    Property TypeClose: Byte Read FTypeClose Write SetTypeClose;
    Property Mode: Byte Read FMode Write FMode;
    Property TextWrap: Integer Read FTextWrap Write FTextWrap;
    Property CharLineLength: Byte Read FMaxLineWidth;
    Property CurrentDeviceNumber: Integer Read FDeviceNumber Write FDeviceNumber;
    Property BufferPDEmpty: Boolean Read FBufferPDEmpty;
    Property BufferPDFull: Boolean Read FBufferPDFull;
    Property BatteryLow: Boolean Read FBatteryLow;
    Property CheckNumber: Integer Read FCheckNumber;
    Property CheckState: Byte Read FCheckState;
    Property CoverOpened: Boolean Read FCoverOpened;
    Property ECRDateTime: TDateTime Read FECRDateTime;
    Property OutOfPaper: Boolean Read FOutOfPaper;
    Property PaperError: Boolean Read FPaperError;
    Property PrinterCutMechanismError: Boolean Read FPrinterCutMechanismError;
    Property PrinterOverheatError: Boolean Read FPrinterOverheatError;
    Property Session: Integer Read FSession;
    Property SessionOpened: Boolean Read FSessionOpened;
    Property Umodel: Byte Read FUmodel;
    Property ReportType: Byte Read FReportType Write FReportType;
    Property TableNumber: Integer Read FTableNumber Write FTableNumber;
    Property FieldNumber: Integer Read FFieldNumber Write FFieldNumber;
    Property RowNumber: Integer Read FRowNumber Write FRowNumber;
    Property FieldType: Boolean Read FFieldType Write Set_FieldType;
    Property DebugFunc: TDrvDebugFunc Read FDebugFunc Write FDebugFunc;
  End;

Implementation
  //****************************************************************************//
  {Служебные методы модуля}
  //****************************************************************************//

Function CRC(s: Ansistring): Byte;
Var
  i: Integer;
  a: Byte;
Begin
  a := 0;
  For i := 1 To Length(s) Do
  Begin
    a := a Xor Byte(s[i]);
  End;
  Result := a;
End;

Function ValueToBCD(Value: DWord; Digits: Integer): String;
{var
 i: integer;
  b: byte;
  s: string;
begin
  SetLength(s, Digits div 2);
  for i:=(Digits div 2) downto 1 do
  begin
    B:=Value mod 10;
    Value:=Value div 10;
    B:=(Value mod 10)* 16 + B;
    Value:=Value div 10;
    s[i]:=Char(B);
  end;
  while Length(s)<Digits do
  begin
    s:=char(0)+s;
  end;
  Result:=s;
end;}
Var
  i:     Integer;
  s, s1: String;
  odd:   Boolean;
  b, b1: Byte;
Begin
  odd := False;
  b   := 0;
  b1  := 0;
  s1  := '';
  s   := IntToStr(Value);
  For i := length(s) Downto 1 Do
  Begin
    If odd = False Then
    Begin
      b   := StrToInt(s[i]);
      odd := True;
    End
    Else
    Begin
      b1  := StrToInt(s[i]);
      s1  := Char((b1 * 16) + b) + s1;
      odd := False;
    End;
  End;
  If odd = True Then
  Begin
    s1 := Char(b) + s1;
  End;
  While Length(s1) < Digits Do
  Begin
    s1 := Char(0) + s1;
  End;
  Result := s1;
End;

Function KKMErrorToStr(ACode: Integer): String;
Begin
  Case ACode Of
    $00: Result := 'Ошибок нет';
    $01: Result := 'Контрольная лента обработана без ошибок';
    $08: Result := 'Неверная цена (сумма)';
    $0A: Result := 'Неверное количество';
    $0B: Result := 'Переполнение счетчика наличности';
    $0C: Result := 'Невозможно сторно последней операции';
    $0D: Result := 'Сторно по коду невозможно (в чеке зарегистрировано меньшее количество товаров с указанным кодом)';
    $0E: Result := 'Невозможен повтор последней операции';
    $0F: Result := 'Повторная скидка на операцию невозможна';
    $10: Result := 'Скидка/надбавка на предыдущую операцию невозможна';
    $11: Result := 'Неверный код товара';
    $12: Result := 'Неверный штрихкод товара';
    $13: Result := 'Неверный формат';
    $14: Result := 'Неверная длина';
    $15: Result := 'ККТ заблокирована в режиме ввода даты';
    $16: Result := 'Требуется подтверждение ввода даты';
    $18: Result := 'Нет больше данных для передачи ПО ККТ';
    $19: Result := 'Нет подтверждения или отмены продажи';
    $1A: Result := 'Отчет с гашением прерван. Вход в режим невозможен.';
    $1B: Result := 'Отключение контроля наличности невозможно (не настроены необходимые типы оплаты).';
    $1E: Result := 'Вход в режим заблокирован';
    $1F: Result := 'Проверьте дату и время';
    $20: Result := 'Дата и время в ККТ меньше чем в ЭКЛЗ/ФП';
    $21: Result := 'Невозможно закрыть архив';
    $3D: Result := 'Товар не найден';
    $3E: Result := 'Весовой штрихкод с количеством <>1.000';
    $3F: Result := 'Переполнение буфера чека';
    $40: Result := 'Недостаточное количество товара';
    $41: Result := 'Сторнируемое количество больше проданного';
    $42: Result := 'Заблокированный товар не найден в буфере чека';
    $43: Result := 'Данный товар не продавался в чеке, сторно невозможно';
    $44: Result := 'Memo PlusTM 3TM заблокировано с ПК';
    $45: Result := 'Ошибка контрольной суммы таблицы настроек Memo PlusTM 3TM';
    $46: Result := 'Неверная команда от ККТ';
    $66: Result := 'Команда не реализуется в данном режиме ККТ';
    $67: Result := 'Нет бумаги';
    $68: Result := 'Нет связи с принтером чеков';
    $69: Result := 'Механическая ошибка печатающего устройства';
    $6A: Result := 'Неверный тип чека';
    $6B: Result := 'Нет больше строк картинки';
    $6C: Result := 'Неверный номер регистра';
    $6D: Result := 'Недопустимое целевое устройство';
    $6E: Result := 'Нет места в массиве картинок';
    $6F: Result := 'Неверный номер картинки / картинка отсутствует';
    $70: Result := 'Сумма сторно больше, чем было получено данным типом оплаты';
    $71: Result := 'Сумма не наличных платежей превышает сумму чека';
    $72: Result := 'Сумма платежей меньше суммы чека';
    $73: Result := 'Накопление меньше суммы возврата или аннулирования';
    $75: Result := 'Переполнение суммы платежей';
    $76: Result := '(зарезервировано)';
    $7A: Result := 'Данная модель ККТ не может выполнить команду';
    $7B: Result := 'Неверная величина скидки / надбавки';
    $7C: Result := 'Операция после скидки / надбавки невозможна';
    $7D: Result := 'Неверная секция';
    $7E: Result := 'Неверный вид оплаты';
    $7F: Result := 'Переполнение при умножении';
    $80: Result := 'Операция запрещена в таблице настроек';
    $81: Result := 'Переполнение итога чека';
    $82: Result := 'Открыт чек аннулирования – операция невозможна';
    $84: Result := 'Переполнение буфера контрольной ленты';
    $86: Result := 'Вносимая клиентом сумма меньше суммы чека';
    $87: Result := 'Открыт чек возврата – операция невозможна';
    $88: Result := 'Смена превысила 24 часа';
    $89: Result := 'Открыт чек продажи – операция невозможна';
    $8A: Result := 'Переполнение ФП';
    $8C: Result := 'Неверный пароль';
    $8D: Result := 'Буфер контрольной ленты не переполнен';
    $8E: Result := 'Идет обработка контрольной ленты';
    $8F: Result := 'Обнуленная касса (повторное гашение невозможно)';
    $91: Result := 'Неверный номер таблицы';
    $92: Result := 'Неверный номер ряда';
    $93: Result := 'Неверный номер поля';
    $94: Result := 'Неверная дата';
    $95: Result := 'Неверное время';
    $96: Result := 'Сумма чека по секции меньше суммы сторно';
    $97: Result := 'Подсчет суммы сдачи невозможен';
    $98: Result := 'В ККТ нет денег для выплаты';
    $9A: Result := 'Чек закрыт – операция невозможна';
    $9B: Result := 'Чек открыт – операция невозможна';
    $9C: Result := 'Смена открыта, операция невозможна';
    $9D: Result := 'ККТ заблокирована, ждет ввода пароля доступа к ФП';
    $9E: Result := 'Заводской номер уже задан';
    $9F: Result := 'Исчерпан лимит перерегистраций';
    $A0: Result := 'Ошибка ФП';
    $A2: Result := 'Неверный номер смены';
    $A3: Result := 'Неверный тип отчета';
    $A4: Result := 'Недопустимый пароль';
    $A5: Result := 'Недопустимый заводской номер ККТ';
    $A6: Result := 'Недопустимый РНМ';
    $A7: Result := 'Недопустимый ИНН';
    $A8: Result := 'ККТ не фискализирована';
    $A9: Result := 'Не задан заводской номер';
    $AA: Result := 'Нет отчетов';
    $AB: Result := 'Режим не активизирован';
    $AC: Result := 'Нет указанного чека в КЛ';
    $AD: Result := 'Нет больше записей КЛ';
    $AE: Result := 'Некорректный код или номер кода защиты ККТ';
    $B0: Result := 'Требуется выполнение общего гашения';
    $B1: Result := 'Команда не разрешена введенными кодами защиты ККТ';
    $B2: Result := 'Невозможна отмена скидки/надбавки';
    $B3: Result := 'Невозможно закрыть чек данным типом оплаты (в чеке присутствуют операции без контроля наличных)';
    $B4: Result := 'Неверный номер маршрута';
    $B5: Result := 'Неверный номер начальной зоны';
    $B6: Result := 'Неверный номер конечной зоны';
    $B7: Result := 'Неверный тип тарифа';
    $B8: Result := 'Неверный тариф';
    $BA: Result := 'Ошибка обмена с фискальным модулем';
    $BE: Result := 'Необходимо провести профилактические работы';
    $BF: Result := 'Неверные номера смен в ККТ и ЭКЛЗ';
    $C8: Result := 'Нет устройства, обрабатывающего данную команду';
    $C9: Result := 'Нет связи с внешним устройством';
    $CA: Result := 'Ошибочное состояние ТРК';
    $CB: Result := 'Больше одной регистрации в чеке';
    $CC: Result := 'Ошибочный номер ТРК';
    $CD: Result := 'Неверный делитель';
    $CF: Result := 'Исчерпан лимит активизаций';
    $D0: Result := 'Активизация данной ЭКЛЗ в составе данной ККТ невозможна';
    $D1: Result := 'Перегрев головки принтера';
    $D2: Result := 'Ошибка обмена с ЭКЛЗ на уровне интерфейса I2C';
    $D3: Result := 'Ошибка формата передачи ЭКЛЗ';
    $D4: Result := 'Неверное состояние ЭКЛЗ';
    $D5: Result := 'Неисправимая ошибка ЭКЛЗ';
    $D6: Result := 'Авария крипто-процессора ЭКЛЗ';
    $D7: Result := 'Исчерпан временной ресурс ЭКЛЗ';
    $D8: Result := 'ЭКЛЗ переполнена';
    $D9: Result := 'В ЭКЛЗ переданы неверная дата или время';
    $DA: Result := 'В ЭКЛЗ нет запрошенных данных';
    $DB: Result := 'Переполнение ЭКЛЗ (итог чека)';
    $DC: Result := 'Буфер переполнен';
    $DD: Result := 'Невозможно напечатать вторую фискальную копию';
    $DE: Result := 'Требуется гашение ЭЖ';
    $DF: Result := 'Сумма налога больше суммы регистраций по чеку и/или итога или больше суммы регистрации';
    $E0: Result := 'Начисление налога на последнюю операцию невозможно';
    $E1: Result := 'Неверный номер ЭКЛЗ';
    $E4: Result := 'Сумма сторно налога больше суммы зарегистрированного налога данного типа';
    $E5: Result := 'Ошибка SD';
    $E6: Result := 'Операция невозможна, недостаточно питания';
    Else
      Result := 'Ошибка не определена';
  End;
End;

Function WrapText(vText: String; vWrapType, vLineWidth: Integer): TStringList;
Var
  ResList: TStringList;
  i, j:    Integer;
Begin
  ResList := TStringList.Create;
  vText   := Trim(vText);
  Case vWrapType Of
    1:
    Begin //Перенос по словам
      While length(vText) > vLineWidth Do
      Begin
        j := vLineWidth;
        For i := 1 To vLineWidth Do
        Begin
          If vText[i] In [' ', '.', ',', ':', ';', '-', '?', '!', ')', '}', ']'] Then
            j := i;
        End;
        ResList.Add(Trim(copy(vText, 1, j)));
        vText := Trim(vText);
        vText := copy(vText, j + 1, length(vText) - j);
      End;
      ResList.Add(trim(vText));
    End;
    2:
    Begin //Перенос по ширине строки
      While length(vText) > vLineWidth Do
      Begin
        ResList.Add(copy(vText, 1, vLineWidth));
        vText := copy(vText, vLineWidth + 1, length(vText) - vLineWidth);
      End;
      ResList.Add(vText);
    End
    Else
    Begin
      If length(vText) > vLineWidth Then
        vText := copy(vText, 1, vLineWidth);
      ResList.Add(vText);
    End;
  End;
  Result := ResList;
End;

  //****************************************************************************//
  {TFRAtolDriver}
  //****************************************************************************//
Constructor TFRAtolDriver.Create;
Begin
  FSerialPort             := TBlockSerial.Create;
  FSerialPort.LinuxLock   := False;
  FSerialPort.RaiseExcept := False;
  FAccessPassword         := 0;
  FPassword               := '30';
  FTypeClose              := 1;
  FTestMode               := False;
  FCheckType              := 1;
  FMaxLineWidth           := 20;
End;

Destructor TFRAtolDriver.Destroy;
Begin
  If Assigned(FSerialPort) Then
  Begin
    FSerialPort.Purge;
    FreeAndNil(FSerialPort);
  End;
End;

Procedure TFRAtolDriver.SetErrorCode(AErrorCode: Integer);
Begin
  FErrorCode    := AErrorCode;
  FErrorCodeStr := KKMErrorToStr(AErrorCode);
End;

Procedure TFRAtolDriver.DebugMsg(msg: String);
Begin
  If Assigned(DebugFunc) Then
    DebugFunc(msg);
End;

Procedure TFRAtolDriver.SetKKMConected(AValue: Boolean);
Begin
  If FKKMConnected <> AValue Then
    If FKKMConnected Then
    Begin
      If Assigned(FSerialPort) Then
      Begin
        FSerialPort.CloseSocket;
      End;
      FKKMConnected := False;
    End
    Else
    Begin
      FErrorCode := 0;
      FSerialPort.Connect(FPortName);
      If FSerialPort.InstanceActive = True Then
      Begin
        FSerialPort.Config(FPortSpeed, FDataBits, FParitySim, StopBits, FXOnXOff, FRTSCTS);
        //FSerialPort.Config(115200, 8, 'N', 0, False, False);
        If FSerialPort.LastError <> sOk Then
        Begin
          FSerialPort.CloseSocket;
          FKKMConnected := False;
          FErrorCode    := -1;
          FErrorCodeStr := FSerialPort.LastErrorDesc;
        End
        Else
        Begin
          FKKMConnected := True;
        End;
      End
      Else
      Begin
        FKKMConnected := False;
        FErrorCode    := -1;
        FErrorCodeStr := 'Не установлена связь -' + FPortName;
      End;
    End;
End;

Procedure TFRAtolDriver.SetDepartment(AValue: Integer);
Begin
  If FDepartment = AValue Then
    Exit;
  If AValue > 30 Then
    AValue    := 30;
  FDepartment := AValue;
End;

Procedure TFRAtolDriver.SetCheckType(AValue: Byte);
Begin
  If FCheckType = AValue Then
    Exit;
  FCheckType := AValue;
End;

Procedure TFRAtolDriver.SetCount(AValue: Integer);
Begin
  If FCount = AValue Then
    Exit;
  FCount := AValue;
End;

Procedure TFRAtolDriver.SetAccessPassword(AValue: Integer);
Begin
  If FAccessPassword = AValue Then
    Exit;
  FAccessPassword := AValue;
End;

Procedure TFRAtolDriver.SetCaption(AValue: String);
Begin
  If AValue <> Caption Then
    FCaption := UTF8ToCP866(AValue);
End;

Function TFRAtolDriver.GetCaption: String;
Begin
  Result := FCaption;
End;

Procedure TFRAtolDriver.SetPassword(AValue: String);
Begin
  If FPassword = AValue Then
    Exit;
  FPassword := AValue;
End;

Procedure TFRAtolDriver.SetPortName(AValue: String);
Begin
  If FPortName = AValue Then
    Exit;
  FPortName := AValue;
End;

Procedure TFRAtolDriver.SetPortSpeed(AValue: Integer);
Begin
  If FPortSpeed = AValue Then
    Exit;
  FPortSpeed := AValue;
End;

Procedure TFRAtolDriver.SetParity(AValue: Char);
Begin
  If FParitySim = AValue Then
    Exit;
  FParitySim := AValue;
End;

Procedure TFRAtolDriver.SetPrice(AValue: Currency);
Begin
  If FPrice = AValue Then
    Exit;
  FPrice := AValue;
End;

Procedure TFRAtolDriver.Set_FieldType(AValue: Boolean);
Begin
  If FFieldType = AValue Then
    Exit;
  FFieldType := AValue;
End;

Procedure TFRAtolDriver.SetQuantity(AValue: Double);
Begin
  If FQuantity = AValue Then
    Exit;
  FQuantity := AValue;
End;

Procedure TFRAtolDriver.SetTestMode(AValue: Boolean);
Begin
  If FTestMode = AValue Then
    Exit;
  FTestMode := AValue;
End;

Procedure TFRAtolDriver.SetTypeClose(AValue: Byte);
Begin
  If FTypeClose = AValue Then
    Exit;
  FTypeClose := AValue;
End;

  //****************************************************************************//
  //****************************************************************************//
Function DeMask(AData: Ansistring): Ansistring;
Var
  i:        Integer;
  DLE_Flag: Boolean;
Begin
  Result   := '';
  DLE_Flag := False;
  For i := 1 To length(AData) Do
  Begin
    If DLE_FLag Then
      DLE_Flag := False
    Else
    If AData[i] = sDLE Then
    Begin
      If i < length(AData) Then
        If AData[i + 1] In [sDLE, sETX] Then
        Begin
          Result   := Result + AData[i + 1];
          DLE_Flag := True;
        End;
    End
    Else
      Result := Result + AData[i];
  End;
End;

Function TFRAtolDriver.MakeCommand(AData: Ansistring): Ansistring;
Var
  i: Integer;
  B: Byte;
Begin
  Result := '';
  AData  := ValueTOBCD(FAccessPassword, 2) + AData;
  //1. Маскируем байты, равные DLE и ETX (10h и 03h):
  For i := 1 To Length(AData) Do
  Begin
    If AData[i] In [sDLE, sETX] Then
      Result := Result + sDLE;
    Result   := Result + AData[i];
  End;
  //2. Добавляем в конец ETX
  Result := Result + sETX;
  //3. Подсчитываем <CRC>
  B      := CRC(Result);
  //4. Добавляем в начало STX
  Result := sSTX + Result;
  //5. Добавляем в конец <CRC>
  Result := Result + Char(B);
End;

Procedure TFRAtolDriver.SendCommand(cmd: Ansistring; TimeOutAnsverCmd: Integer = timeout_T5);
Label
  l1, l2, l3, l4, l5, l8, l9, l10,
  l11, l12, l13, l14, l15, l16;
Var
  FRC, RC: Integer;
  B:       Char;
  S:       String;
Begin
  FRC  := 0;
  l1:
    RC := 0;
  FRC  := FRC + 1;
  l2:
    RC := RC + 1;
  If RC <= 5 Then
    Goto l3
  Else
    Goto l9;
  l3:
    //Посылаем ENQ
    If FSerialPort.InstanceActive = True Then
    Begin
      S := sENQ;
      FSerialPort.SendBuffer(@S[1], Length(S));
      Sleep(1);
    End
    Else
    Begin
      FKKMConnected := False;
      FErrorCode    := -1;
      FErrorCodeStr := 'Не установлена связь -' + FPortName;
      exit;
    End;
  // Ждем ACK
  If FSerialPort.CanRead(timeout_T1) Then
  Begin
    FSerialPort.RecvBuffer(@B, 1);
    If B = sNAK Then
      Goto l4
    Else If B = sENQ Then
      Goto l5
    Else If B = sACK Then
      Goto l10
    Else
      Goto l8;
  End
  Else
    Goto l2;
  l4:
    Sleep(timeout_T1);
  Goto l2;

  l5:
    Sleep(timeout_T7);
  l8:
    If FRC <= 100 Then
      Goto l1;
  l9:
    //Посылаем EOT
    If FSerialPort.InstanceActive = True Then
    Begin
      S := sEOT;
      FSerialPort.SendBuffer(@S[1], Length(S));
      Sleep(1);
    End
    Else
    Begin
      FKKMConnected := False;
      FErrorCode    := -1;
      FErrorCodeStr := 'Не установлена связь -' + FPortName;
      exit;
    End;
  Goto l15;
  l10:
    RC := 0;
  l11:
    RC := RC + 1;
  If RC <= 10 Then
    Goto l12
  Else
    Goto l9;
  l12:
    // Посылаем cmd
    If FSerialPort.InstanceActive = True Then
    Begin
      FSerialPort.SendBuffer(@cmd[1], Length(cmd));
      Sleep(1);
    End
    Else
    Begin
      FKKMConnected := False;
      FErrorCode    := -1;
      FErrorCodeStr := 'Не установлена связь -' + FPortName;
      exit;
    End;
  // Ждем ACK
  If FSerialPort.CanRead(timeout_T3) Then
  Begin
    FSerialPort.RecvBuffer(@B, 1);
    If (B = sENQ) And (RC > 1) Then
      exit
    Else If B = sACK Then
      Goto l13
    Else
      Goto l11;
  End
  Else
    Goto l11;
  l13:
    //Посылаем EOT
    If FSerialPort.InstanceActive = True Then
    Begin
      S := sEOT;
      FSerialPort.SendBuffer(@S[1], Length(S));
    End
    Else
    Begin
      FKKMConnected := False;
      FErrorCode    := -1;
      FErrorCodeStr := 'Не установлена связь -' + FPortName;
      exit;
    End;
  RC   := 0;
  l14:
    RC := RC + 1;
  If RC <= 100 Then
    Goto l16
  Else
    Goto l15;
  l15:
    FErrorCode  := -1;
  FErrorCodeStr := 'Нет ответа ККМ';
  exit;
  l16:
    //Ждем ответ в течении TimeOutAnsverCmd
    If FSerialPort.CanRead(TimeOutAnsverCmd) Then
    Begin
      FSerialPort.RecvBuffer(@B, 1);
      If B <> sENQ Then
        Goto l14;
    End
    Else
      Goto l15;
End;

Function TFRAtolDriver.ReadAnsver: Ansistring;
Label
  l1, l2, l3, l4, l5, l6, l7, l8, l9, l10,
  l11, l12, {l13,} l14, l15, l16, l17;
Var
  FRC, RC, BC: Integer;
  B:           Char;
  DLE_Flag:    Integer;
  S:           Ansistring;
  //sCRC:        Byte;
Begin
  Result := '';
  BC     := 0;
  FRC    := 0;
  l1:
    //Посылаем ACK
    If FSerialPort.InstanceActive = True Then
    Begin
      S := sACK;
      FSerialPort.SendBuffer(@S[1], Length(S));
    End
    Else
    Begin
      FKKMConnected := False;
      FErrorCode    := -1;
      FErrorCodeStr := 'Не установлена связь -' + FPortName;
      Result        := '';
      exit;
    End;
  l2:
    RC := 0;
  l3:
    RC := RC + 1;
  If RC <= 100 Then
    Goto l5
  Else
    Goto l4;
  l4:
    FErrorCode  := -1;
  FErrorCodeStr := 'Нет ответа ККМ';
  Result        := '';
  exit;
  l5:
    //Ждем STX
    If FSerialPort.CanRead(timeout_T2) Then
    Begin
      FSerialPort.RecvBuffer(@B, 1);
      If B = sENQ Then
        Goto l6
      Else If B = sSTX Then
        Goto l7
      Else
        Goto l3;
    End
    Else
      Goto l4;
  l6:
    FRC := FRC + 1;
  If FRC <= 10 Then
    Goto l1
  Else
    Goto l4;
  l7:
    BC     := 0;
  DLE_Flag := 0;
  l8:
    If BC > BMax Then
      Goto l12;
  //Читаем байт ответа ККМ
  If FSerialPort.CanRead(timeout_T6) Then
  Begin
    FSerialPort.RecvBuffer(@B, 1);
  End
  Else
    Goto l12;
  If DLE_Flag = 1 Then
    Goto l11;
  If B <> sDLE Then
    Goto l10;
  DLE_Flag := 1;
  l9:
    Result := Result + B;
  BC       := BC + 1;
  Goto l8;
  l10:
    If B = sETX Then
      Goto l14
    Else
      Goto l9;
  l11:
    DLE_Flag := 0;
  Goto l9;
  l12:
    FRC := FRC + 1;
  If FRC <= 10 Then
    Goto l2
  Else
    Goto l4;
  //l13:
  //  FRC := FRC + 1;
  //If FRC <= 10 Then
  //  Goto l2
  //Else
  //  Goto l4;
  l14:
    //Читаем байт ответа ККМ
    If FSerialPort.CanRead(timeout_T6) Then
    Begin
      FSerialPort.RecvBuffer(@B, 1);
    End
    Else
    Begin
      Sleep(1);
      Goto l12;
    End;
  //проверка crc
  //sCRC:=CRC(Result+sETX);
  //if CRC(Result+sETX)<>Byte(B) then
  //begin
  //  S := sNAK;
  //  FSerialPort.SendBuffer(@S[1], Length(S));
  //  goto l12;
  //end;
  //Посылаем ACK
  If FSerialPort.InstanceActive = True Then
  Begin
    S := sACK;
    FSerialPort.SendBuffer(@S[1], Length(S));
  End
  Else
  Begin
    FKKMConnected := False;
    FErrorCode    := -1;
    FErrorCodeStr := 'Не установлена связь -' + FPortName;
    Result        := '';
    exit;
  End;
  //Ждем EOT или STX
  If FSerialPort.CanRead(timeout_T4) Then
  Begin
    FSerialPort.RecvBuffer(@B, 1);
    If B = sEOT Then
      Goto l15
    Else If B = sSTX Then
      Goto l16
    Else
      Goto l17;
  End
  Else
    Goto l15;
  l15:
    Exit;
  l16:
    FRC := FRC + 1;
  If FRC <= 10 Then
    Goto l4
  Else
    Goto l7;
  l17:
    If FSerialPort.CanRead(timeout_T6) Then
    Begin
      FSerialPort.RecvBuffer(@B, 1);
      Goto l12;
    End
    Else
      Goto l15;
End;

Procedure TFRAtolDriver.InternalGetStatus;
Var
  S, SR: Ansistring;
Begin
  FStatus  := 0;
  FMode    := 0;
  FSubMode := 0;

  S := MakeCommand(#$45); //Запрос кода состояния ККМ
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) >= 3 Then
  Begin
    FStatus                   := Byte(SR[3]);
    FMode                     := Byte(SR[2]) Mod 16;
    FSubMode                  := Byte(SR[2]) Div 16;
    FOutOfPaper               := (FStatus And $01) <> 0;
    FPrintConected            := (FStatus And $02) = 0;
    FPaperError               := (FStatus And $04) <> 0;
    FPrinterCutMechanismError := (FStatus And $08) <> 0;
    FPrinterOverheatError     := (FStatus And $10) <> 0;
    FBufferPDEmpty            := (FStatus And $20) <> 0;
    FBufferPDFull             := (FStatus And $40) <> 0;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.InternalSetMode(AMode: Byte);
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$56 + Char(AMode) + ValueToBCD(StrToInt(FPassword), 4)); //Вход в режим
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

  //****************************************************************************//
  //****************************************************************************//
Procedure TFRAtolDriver.Beep;
Var
  S: String;
Begin
  S := MakeCommand(#$47);
  SendCommand(S, 1);
End;

Procedure TFRAtolDriver.FullCut;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$75#$00);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.PartialCut;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$75#$01);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.PrintString;
Var
  S, SR: Ansistring;
  vSL:   TStringList;
  i:     Integer;
Begin
  If FCaption <> '' Then
  Begin
    vSL := WrapText(FCaption, FTextWrap, FMaxLineWidth);
    For i := 0 To vSL.Count - 1 Do
    Begin
      S := MakeCommand(#$4C + vSL[i]);
      SendCommand(S, timeout_T5);
      SR := DeMask(ReadAnsver);
      If Length(SR) = 3 Then
      Begin
        FErrorCode    := Byte(SR[2]);
        FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
      End
      Else
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Нет ответа ККМ';
      End;
    End;
    FreeAndNil(vSL);
  End;
End;

Procedure TFRAtolDriver.PrintHeader;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$6C);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.CancelCheck;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$59);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.Payment;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$99 + Char(FTestMode) + Char(FTypeClose + 1) + ValueToBCD(trunc(FSumm * 100), 5));
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.OpenDrawer;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$80);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.OpenCheck;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$92 + Char(FTestMode) + Char(FCheckType));
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.ResetMode;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$48);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.GetStatus;
Var
  S, SR:              Ansistring;
  Flags:              Byte;
  y, m, d, h, mm, ss: Integer;
Begin
  FMode          := 0;
  FSubMode       := 0;
  FCheckNumber   := 0;
  FCheckState    := 0;
  FSession       := 0;
  FSessionOpened := False;
  FOutOfPaper    := False;
  FCoverOpened   := False;
  FBatteryLow    := False;

  S := MakeCommand(#$3F);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  {"D"<Кассир(1)> <Номер_в_зале(1)> <Дата_YMD(3)>
<Время_HMS(3)> <Флаги(1)> <Заводской_номер(4)> <Модель(1)>
<Версия_ПО_ККМ(2)> <Режим_работы(1)> <Номер_чека(2)>
<Номер_смены(2)> <Состояние_чека(1)> <Сумма_чека(5)>
<Десятичная_точка(1)> <Порт(1)> }
y :=Length(SR);
  If Length(SR) >= 30 Then
  Begin
    y := BCDToInt(Byte(SR[4])); //<Дата_YMD(3)>
    m := BCDToInt(Byte(SR[5])); //<Дата_YMD(3)>
    d := BCDToInt(Byte(SR[6])); //<Дата_YMD(3)>
    If y < 90 Then
      y := 2000 + y
    Else
      y := 1900 + y;
    h := BCDToInt(Byte(SR[7]));               //<Время_HMS(3)>
    mm             := BCDToInt(Byte(SR[8]));  //<Время_HMS(3)>
    ss             := BCDToInt(Byte(SR[9]));  //<Время_HMS(3)>
    FECRDateTime   := EncodeDate(y, m, d) + EncodeTime(h, mm, ss, 0);
    Flags          := Byte(SR[10]);
    // 11, 12, 13, 14 - заводской номер
    // 15 - модель
    // 16, 17 - Версия ПО
    FMode          := Byte(SR[18]) Mod 16;
    FSubMode       := Byte(SR[18]) Div 16;
    FCheckNumber   := BCDToInt(Byte(SR[19]) * 256 + Byte(SR[20]));
    FCheckState    := Byte(SR[23]);
    FSession       := BCDToInt(Byte(SR[21]) * 256 + Byte(SR[22]));
    FSessionOpened := Flags And 2 <> 0;
    FOutOfPaper    := Flags And 8 <> 0;
    FCoverOpened   := Flags And 32 <> 0;
    FBatteryLow    := Flags And 128 <> 0;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.GetDeviceMetrics;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$A5);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  {<Код_ошибки (1)>, <Версия_протокола (1)>, <Тип (1)>, <Модель (1)>,
   <Режим (2)>, <Версия_устройства (5)>, <Название (N)>}
  If length(SR) >= 11 Then
  Begin
    FErrorCode    := Byte(SR[1]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[1]));
    FUmodel       := Byte(SR[4]);
    Case FUmodel Of
      13: //Триум-Ф
        FMaxLineWidth := 40;
      14: //ФЕЛИКС-Р Ф
        FMaxLineWidth := 20;
      15: //ФЕЛИКС-02К;
        FMaxLineWidth := 20;
      16: //«Меркурий-140Ф» АТОЛ
        FMaxLineWidth := 24;
      20: //ТОРНАДО
        FMaxLineWidth := 48;
      23: //Меркурий MS-K
        FMaxLineWidth := 39;
      24: //ФЕЛИКС-Р К
        FMaxLineWidth := 38;
      27: //ФЕЛИКС-3СК
        FMaxLineWidth := 38;
      30: //FPrint-02K
        FMaxLineWidth := 56;
      31: //FPrint-03K
        FMaxLineWidth := 32;
      32: //FPrint-88K
        FMaxLineWidth := 56;
      35: //FPrint-5200K
        FMaxLineWidth := 36;
      41: //PayVKP-80K
        FMaxLineWidth := 56;
      45: //PayPPU-700K
        FMaxLineWidth := 56;
      46: //PayCTS-2000K
        FMaxLineWidth := 72;
      47: //FPrint-55K
        FMaxLineWidth := 36;
      52: //FPrint-22K
        FMaxLineWidth := 48;
    End;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.Registration;
Var
  S, SR: Ansistring;
Begin
  If (FPrice <> 0) And (FQuantity <> 0) Then
  Begin
    PrintString;
    If Not FErrorCode <> 0 Then
    Begin
      S := MakeCommand(#$52 + Char(FTestMode) + ValueToBCD(trunc(FPrice * 100), 5) + ValueToBCD(trunc(FQuantity * 1000), 5) +
        ValueToBCD(FDepartment, 1));
      SendCommand(S, timeout_T5);
      SR := DeMask(ReadAnsver);
      {"U"<Код Ошибки(1)><0>}
      If Length(SR) = 3 Then
      Begin
        FErrorCode    := Byte(SR[2]);
        FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
      End
      Else
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Нет ответа ККМ';
      End;
    End;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Регистрация: Цена или Количество не могут быть нулевыми';
  End;
End;

Procedure TFRAtolDriver.Return;
Var
  S, SR: Ansistring;
Begin
  If (FPrice <> 0) And (FQuantity <> 0) Then
  Begin
    PrintString;
    //"W" <Флаги(1)><Цена(5)><Количество(5)>
    S := MakeCommand(#$57 + Char(0) + ValueToBCD(trunc(FPrice * 100), 5) + ValueToBCD(trunc(FQuantity * 1000), 5));
    SendCommand(S, timeout_T5);
    SR := DeMask(ReadAnsver);
    {"U"<Код Ошибки(1)><0>}
    If Length(SR) = 3 Then
    Begin
      FErrorCode    := Byte(SR[2]);
      FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
    End
    Else
    Begin
      FErrorCode    := -1;
      FErrorCodeStr := 'Нет ответа ККМ';
    End;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Регистрация: Цена или Количество не могут быть нулевыми';
  End;
End;

Procedure TFRAtolDriver.CashIncome;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$49 + Char(FTestMode) + ValueToBCD(trunc(FSumm * 100), 5));
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.CashOutcome;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$4F + Char(FTestMode) + ValueToBCD(trunc(FSumm * 100), 5));
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.OpenSession;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$9A + Char(FTestMode) + FCaption);
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.CloseCheck;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$4A + Char(FTestMode) + Char(FTypeClose) + ValueToBCD(trunc(0 * 100), 5));
  SendCommand(S, 20000);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.ReportX;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$67 + Char(FReportType));
  SendCommand(S, timeout_T5);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.ReportZ;
Var
  S, SR: Ansistring;
Begin
  S := MakeCommand(#$5A); //Начало отчета с гашением
  SendCommand(S, 40000);
  SR := DeMask(ReadAnsver);
  If Length(SR) = 3 Then
  Begin
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;

  //Дожидаемся окончания отчета с гашением
  If FErrorCode = 0 Then
  Begin
    Repeat
      InternalGetStatus;
      sleep(500);
    Until (FMode = 3) And (FSubMode = 2);
    If ((FMode <> 7) And (FSubMode <> 1)) And ((FMode <> 3) And (FSubMode <> 2)) Then
    Begin
      //Проверяем состояние процесса формрования отчета с гашением
      {если бит0 поля Флаги= 1, то ошибка «Нет бумаги» (на остатке ленты ККМ автоматически печатается «Чек аннулирован» и отчет прерывается),
       иначе если бит1 поля Флаги= 1, то ошибка «Нет связи с принтером чека»,
             иначе(биты0 и1 поля Флаги= 0) ошибка «Снятие отчета прервалось»,
       иначе если бит2 поля Флаги= 1, то ошибка «Механическая ошибка печатающего устройства»,
             иначе(биты0, 1 и2 поля Флаги= 0) ошибка «Снятие отчета прервалось»}
      If (FStatus And $01) = 1 Then
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Нет бумаги. Чек аннулирован.';
      End
      Else
      If (FStatus And $02) = 1 Then
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Нет связи с принтером чека.';
      End
      Else If ((FStatus And $01) = 0) And ((FStatus And $02) = 0) Then
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Снятие отчета прервалось.';
      End
      Else
      If ((FStatus And $03) = 1) And ((FStatus And $02) = 0) Then
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Механическая ошибка печатающего устройства.';
      End
      Else If ((FStatus And $01) = 0) And ((FStatus And $02) = 0) And ((FStatus And $03) = 0) Then
      Begin
        FErrorCode    := -1;
        FErrorCodeStr := 'Снятие отчета прервалось.';
      End;
    End;
    //Ждем выхода из режима отчета с гашением
    While (FMode = 7) And (FSubMode = 1) Do
    Begin
      sleep(500);
      InternalGetStatus;
    End;
  End;
End;

Function TFRAtolDriver.SetMode(AValue: Byte): Integer;
Begin
  FErrorCode    := 0;
  FErrorCodeStr := '';

  InternalGetStatus;
  ResetMode;
  InternalGetStatus;
  If FMode <> 0 Then
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := Format('Ошибка выхода из режима (%d) - %s.', [FMode, AtolModeNames[FMode]]);
  End
  Else
  If AValue <> 0 Then
  Begin
    InternalSetMode(AValue);
    If FErrorCode = 0 Then
    Begin
      //If AValue <> 3 Then
      GetStatus;
      //Else
      //  //InternalGetStatus;
      //  FMode := AValue;
    End;
  End;
  Result := FErrorCode;
End;

Procedure TFRAtolDriver.NewDocument;
Var
  NewMode: Byte;
Begin
  //Проверим состояние ККМ
  NewMode    := FMode;
  FErrorCode := 0;
  GetStatus;
  If FErrorCode <> 0 Then
    exit;
  //Если чек открыт, то отменить его
  If FCheckState > 0 Then
  Begin
    CancelCheck;
    If FErrorCode <> 0 Then
      exit;
  End;
  //Пробуем сменить режим
  FErrorCode := SetMode(NewMode);
End;

Procedure TFRAtolDriver.Report;
Begin
  Case FMode Of
    2: ReportX;
    3: ReportZ;
    Else
    Begin
      FErrorCode    := -1;
      FErrorCodeStr := 'Нет ответа ККМ';
    End;
  End;
End;

Procedure TFRAtolDriver.ReadTable;
Var
  S, SR:             Ansistring;
  i, FieldSize, val: Integer;
  j:                 Ansistring;
Begin
  {Команда: «F» <Таблица(1)><Ряд(2)><Поле(1)>.
   Ответ: «U» <Код Ошибки(1)><Значение(Х)>.
            Код команды («F», 46h, 70).
            Таблица– двоичное число, номер таблицы в ККМ.
            Ряд– двоичное  число(сначала  старший  байт,  потом  младший),
            номер ряда в указанной таблице.
            Поле– двоичное число, номер поля в ряду таблицы}
  S := MakeCommand(#$46 + Char(FTableNumber) + ValueToBCD(FRowNumber, 2) + Char(FFieldNumber));
  SendCommand(S, timeout_T2);
  SR := DeMask(ReadAnsver);
  If Length(SR) > 2 Then
  Begin
    FCaption      := '';
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
    If FErrorCode = 0 Then
    Begin
      If FFieldType Then
      Begin
        For i := 3 To FMaxLineWidth + 2 Do
        Begin
          Try
            val := Byte(SR[i]);
            j   := IntToStr(val);
          Except
            j := '220';
          End;
          FCaption := FCaption + KKM_CODEPAGE[StrToInt(j)];
        End;
      End
      Else
      Begin
        val := 0;
        j   := '';
        Try
          If FTableNumber = 3 Then
            FieldSize := 4
          Else
          If FTableNumber = 2 Then
            FieldSize := 1;
          If (FTableNumber = 2) And (FFieldNumber = 18) Then
          Begin
            FieldSize := 0;
            If Byte(SR[3]) And 4 <> 0 Then
              val := 1
            Else
              val := 0;
          End;
          If FieldSize = 1 Then
            Val := BCDToInt(Byte(SR[3]))
          Else
          If FieldSize = 2 Then
            val := BCDToInt(Byte(SR[3]) * 256 + Byte(SR[4]))
          Else
          If FieldSize = 3 Then
            val := BCDToInt(Byte(SR[3]) * 256 * 256 + Byte(SR[4]) * 256 + Byte(SR[5]))
          Else
          If FieldSize = 4 Then
            val := BCDToInt(Byte(SR[3]) * 256 * 256 * 256 + Byte(SR[4]) * 256 * 256 + Byte(SR[5]) * 256 + Byte(SR[6]));

          If FieldSize <> 0 Then
            val    := BCDToInt(val);
          FCaption := IntToStr(val);
        Except
          FCaption := '';
        End;
      End;
    End;
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

Procedure TFRAtolDriver.WriteTable;
Var
  S, SR: Ansistring;
Begin
  {Команда: «P» <Таблица(1)><Ряд(2)><Поле(1)><Значение(Х)>.
   Ответ: «U» <Код Ошибки(1)><0>.
          Код команды («P», 50h, 80).
          Таблица – двоичное число, номер таблицы в ККМ.
          Ряд– двоичное  число(сначала  старший  байт,  потом  младший),
          номер ряда в указанной таблице.
          Поле– двоичное число, номер поля в ряду таблицы.}
  S := MakeCommand(#$50 + Char(FTableNumber) + ValueToBCD(FRowNumber, 2) + Char(FFieldNumber) + FCaption);
  SendCommand(S, timeout_T2);
  SR := DeMask(ReadAnsver);
  If Length(SR) > 2 Then
  Begin
    FCaption      := '';
    FErrorCode    := Byte(SR[2]);
    FErrorCodeStr := KKMErrorToStr(Byte(SR[2]));
  End
  Else
  Begin
    FErrorCode    := -1;
    FErrorCodeStr := 'Нет ответа ККМ';
  End;
End;

End.
